# coding=utf-8
from __future__ import unicode_literals

import logging

try:
    from django.conf import settings as django_settings
except ImportError:
    logging.warning("Django not installed")
    django_settings = None


SELECTEL_STORAGE = {
    "USER": None,
    "PASSWORD": None,
    "DOMAINS": {},
    "OVERWRITE_FILES": False,
    "USE_GZ": False,
    "AUTH_URL": "https://auth.selcdn.ru/",
    "API_THRESHOLD": 30 * 60,
    "API_MAX_RETRY": 3,
    "API_RETRY_DELAY": 0.1
}

if hasattr(django_settings, "SELECTEL_STORAGE"):
    SELECTEL_STORAGE.update(django_settings.SELECTEL_STORAGE)



# coding=utf-8
from __future__ import unicode_literals

from django import apps as django_apps


class SelectelCDNConfig(django_apps.AppConfig):
    name = "django_selectel"
    verbose_name = "Selectel CDN"



# coding=utf-8
from __future__ import unicode_literals

import gzip
import os
from cStringIO import StringIO

from django.core.files import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from django.utils.functional import cached_property
from django_selectel import settings
from django_selectel.api import SelectelCDNApi


class ApiStorageException(Exception):
    pass


@deconstructible
class ApiStorage(Storage):

    def __init__(self, user=None, password=None, domains=None, overwrite_files=None, use_gz=None):

        self.user = user or settings.SELECTEL_STORAGE.get("USER")

        if not self.user:
            raise ApiStorageException('The "USER" parameter in the SELECTEL_STORAGE settings is not passed or set')

        self.password = password or settings.SELECTEL_STORAGE.get('PASSWORD')

        if not self.password:
            raise ApiStorageException('The "PASSWORD" parameter in the SELECTEL_STORAGE settings is not passed or set')

        self.domains = domains if domains is not None else settings.SELECTEL_STORAGE.get('DOMAINS', {})

        self.overwrite_files = overwrite_files if overwrite_files is not None else settings.SELECTEL_STORAGE.get("OVERRIDE_FILES", False)
        self.use_gz = use_gz if use_gz is not None else settings.SELECTEL_STORAGE.get("USE_GZ", False)

        self._api = SelectelCDNApi(
            user=self.user,
            password=self.password,
            auth_url=settings.SELECTEL_STORAGE['AUTH_URL'],
            threshold=settings.SELECTEL_STORAGE['API_THRESHOLD'],
            max_retry=settings.SELECTEL_STORAGE['API_MAX_RETRY'],
            retry_delay=settings.SELECTEL_STORAGE['API_RETRY_DELAY']
        )

    def get_available_name(self, name, max_length=None):
        if not self.overwrite_files:
            return super(ApiStorage, self).get_available_name(name, max_length)
        return name

    def _parse_path(self, path):
        splited_path = path.split(os.path.sep)
        if len(splited_path) > 1:
            return splited_path[0], os.path.sep.join(splited_path[1:])
        return splited_path[0], ""

    def path(self, name):
        return name

    def exists(self, name):
        container, path = self._parse_path(name)
        return self._api.exist(container, path)

    def size(self, name):
        container, path = self._parse_path(name)
        return self._api.size(container, path)

    def url(self, name):
        container, path = self._parse_path(name)
        print container
        if settings.SELECTEL_STORAGE.get('DOMAINS', {}).get(container):
            return os.path.join(settings.SELECTEL_STORAGE["DOMAINS"][container], path)
        return self._api.get_url(container, path)

    def delete(self, name):
        container, path = self._parse_path(name)
        self._api.remove(container, path)

    def _save(self, name, content):
        container, path = self._parse_path(name)
        if hasattr(content.file, 'seek'):
            content.file.seek(0)

        if self.use_gz:
            g_file = StringIO()
            g_file_gzip = gzip.GzipFile(fileobj=g_file, mode="wb")
            g_file_gzip.write(content.read())
            g_file_gzip.close()
            file_content = g_file.getvalue()
        else:
            file_content = content.read()

        self._api.put(container, path, file_content)
        return name

    def _open(self, name, mode='rb'):
        return SelectelCDNFile(self, name)

    def _read(self, name):
        container, path = self._parse_path(name)
        content = self._api.get(container, path)
        return content


class SelectelCDNFile(File):

    def __init__(self, storage, path):
        self._storage = storage
        self._path = path
        self._content = None
        self._file = None
        self._is_dirty = False

    @cached_property
    def content(self):
        content = self.read()
        self.seek(0)
        return content

    def open(self, mode=None):
        self.seek(0)

    @property
    def name(self):
        return os.path.basename(self._path)

    @cached_property
    def size(self):
        return self._storage.size(self._path)

    @property
    def file(self):
        if not self._file:
            content = self._storage._read(self._path)
            if self._storage.use_gz:
                content_obj = StringIO(content)
                content_obj.seek(0)
                content = gzip.GzipFile(fileobj=content_obj, mode='rb').read()
            self._file = StringIO(content)
            self._file.seek(0)
        return self._file

    def readlines(self):
        return self.file.readlines()

    def read(self, num_bytes=None):
        if num_bytes:
            return self.file.read(num_bytes)
        else:
            return self.file.read()

    def write(self, content):
        self.file.write(content)
        self._is_dirty = True

    def close(self):
        if self._is_dirty:
            self._storage.save(self.name, self)
        self.file.close()



# coding=utf-8

import os
import re

from setuptools import setup

try:
    from pypandoc import convert

    def read_md(f):
        return convert(f, 'rst')
except ImportError:
    def read_md(f):
        return open(f, 'r').read()


def get_packages(package):
    return [dirpath.replace(os.path.sep, ".")
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]
    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    print filepaths
    return {package: filepaths}


def get_version(package):
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('django_selectel')


setup(
    name='django_selectel',
    description="The file store in the Selectel cloud store for Django",
    long_description=read_md('README.md'),
    version=version,
    download_url='https://github.com/KokocGroup/django_selectel/tarball/v{}'.format(version),
    license='BSD',
    url='https://github.com/KokocGroup/django_selectel',
    packages=get_packages('django_selectel'),
    author='KokocGroup',
    author_email='dev@kokoc.com',
    install_requires=[
        'requests==2.18.4'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



# coding=utf-8
from __future__ import unicode_literals

import hashlib
import logging
import os
import time

from datetime import datetime
from datetime import timedelta
from functools import wraps

import requests


class SelectelCDNApiException(Exception):

    def __init__(self, message, *args, **kwargs):
        message = "Selectel: {}".format(message)
        self.response = kwargs.pop('response', None)
        super(SelectelCDNApiException, self).__init__(message)


class SelectelCDNApi(object):

    def __init__(self, user, password, auth_url, threshold=None, max_retry=None, retry_delay=None):
        self.user = user
        self.password = password
        self.auth_url = auth_url
        self.threshold = threshold or 0
        self.max_retry = max_retry
        self.retry_delay = retry_delay

        self._token_expire_dt = None
        self._storage_url = None
        self._session = None
        self.logger = logging.getLogger("SelectelApi")

    def update_expired_token(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self.is_token_expire:
                self.authenticate()
            try:
                return fn(self, *args, **kwargs)
            except (requests.exceptions.HTTPError, SelectelCDNApiException) as e:
                if e.response.status_code == 401:
                    self.authenticate()
                    return fn(self, *args, **kwargs)
                else:
                    raise e

        return wrapper

    def attempts(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self.max_retry is not None:
                retries = self.max_retry
                while retries > 1:
                    try:
                        return fn(self, *args, **kwargs)
                    except (requests.exceptions.HTTPError, SelectelCDNApiException):
                        retries -= 1
                        time.sleep(self.retry_delay)
            return fn(self, *args, **kwargs)

        return wrapper

    @property
    def is_token_expire(self):
        if not self._token_expire_dt:
            return True

        return (self._token_expire_dt - datetime.now()).total_seconds() < self.threshold

    @update_expired_token
    def get_url(self, container, path):
        return os.path.join(self._storage_url, container, path)

    def authenticate(self):
        if not self.user or not self.password:
            raise SelectelCDNApiException("Not set user or password")
        headers = {
            "X-Auth-User": self.user,
            "X-Auth-Key": self.password
        }
        r = requests.get(self.auth_url, headers=headers, verify=True)
        if r.status_code != 204:
            raise SelectelCDNApiException("Authenticate error ({})".format(r.status_code))

        self._token_expire_dt = datetime.now() + timedelta(seconds=int(r.headers['X-Expire-Auth-Token']))
        self._storage_url = r.headers['X-Storage-Url'][:-1]

        self._session = requests.Session()
        self._session.headers.update({
            "X-Auth-Token": r.headers['X-Auth-Token']
        })

    @attempts
    @update_expired_token
    def get(self, container, path, headers=None):
        url = os.path.join(self._storage_url, container, path)
        if headers is None:
            headers = {}
        response = self._session.get(url, headers=headers, verify=True)

        self.logger.info("Request GET {} - {}".format(url, response.status_code))
        try:
            response.raise_for_status()
        except Exception as e:
            raise SelectelCDNApiException("Error get file {}: {}".format(url, str(e)), response=response)
        return response.content

    @attempts
    @update_expired_token
    def get_steam(self, container, path, headers=None, chunk=2 ** 20):
        url = os.path.join(self._storage_url, container, path)
        if headers is None:
            headers = {}
        r = self._session.get(url, headers=headers, stream=True, verify=True)
        self.logger.info("Request GET_STEAM {} - {}".format(url, r.status_code))
        try:
            r.raise_for_status()
        except Exception as e:
            raise SelectelCDNApiException("Error get file {}: {}".format(url, str(e)), response=r)
        return r.iter_content(chunk_size=chunk)

    @attempts
    @update_expired_token
    def remove(self, container, path, force=False):
        url = os.path.join(self._storage_url, container, path)
        r = self._session.delete(url, verify=True)
        self.logger.info("Request REMOVE {} - {}".format(url, r.status_code))
        if force:
            if r.status_code == 404:
                return r.headers
        try:
            r.raise_for_status()
            assert r.status_code == 204
        except Exception as e:
            raise SelectelCDNApiException("Error remove file {}: {}".format(url, str(e)), response=r)
        return r.headers

    @attempts
    @update_expired_token
    def put(self, container, path, content, headers=None):
        url = os.path.join(self._storage_url, container, path)
        if headers is None:
            headers = {}

        headers["ETag"] = hashlib.md5(content).hexdigest()
        r = self._session.put(url, data=content, headers=headers, verify=True)
        self.logger.info("Request PUT {} - {}: {}".format(url, r.status_code, r.content))
        try:
            r.raise_for_status()
            assert r.status_code == 201
        except Exception as e:
            raise SelectelCDNApiException("Error create file {}: {}".format(url, str(e)), response=r)
        return True

    @attempts
    @update_expired_token
    def exist(self, container, path):
        url = os.path.join(self._storage_url, container, path)

        r = self._session.head(url)
        self.logger.info("Request EXIST {} - {}: {}".format(url, r.status_code, r.content))
        if r.status_code not in (200, 404):
            r.raise_for_status()
        return r.status_code == 200

    @attempts
    @update_expired_token
    def size(self, container, path):
        url = os.path.join(self._storage_url, container, path)
        r = self._session.head(url)
        self.logger.info("Request SIZE {} - {}: {}".format(url, r.status_code, r.content))
        r.raise_for_status()
        if r.status_code != 200:
            raise SelectelCDNApiException("file {} not exists".format(os.path.sep.join([container, path])), response=r)
        return int(r.headers['Content-Length'])
